# -*- coding: utf-8 -*-
"""
A Python Class creating a standard "bipedal" Spring-Loaded Inverted
Pendulum (SLIP) Object to model bipedal walking.

The code is adopted from "Spring-Loaded Inverted Pendulum Simulation, by 
Sam Burden, Berkeley 2010". Amongst other changes, the __call__ method is now using 
scipy standard integrator function.

The derivation of this map has been presented in 
"Approximate analytical solutions to the double-stance dynamics of 
the lossy Spring-Loaded Inverted Pendulum, 
by M. Shahbazi, U. Saranli, R. Babuska, G. Lopes" (under review).

Run "test.py" to see sample results.

Copyright (c) 2015 by Mohammad Shahbazi. 
doi:10.4121/uuid:fd95d86d-fa01-4faa-b6e5-eeb92345be8a
http://doi.org/10.4121/uuid:fd95d86d-fa01-4faa-b6e5-eeb92345be8a

This code can be freely used for any ends. 
If use of this code in part or in whole results in publication, proper citation must be
included in that publication. This code comes with no guarantees or support.

Mohammad Shahbazi, August 11, 2015, Delft, the Netherlands.
"""


import numpy as np
import scipy.integrate as integrate
fprint = False

class DBSLIP(object):
    def __init__(self, t, x, q, xdes, dt):
        """
        This is a bipedal SLIP class
        """        
        self.name = 'Bipedal SLIP'
        
        self.t = np.array([t])
        self.x = np.array(x)
        self.q = np.array(q)
        
        self.t0 = self.t[-1]
        self.x0 = self.x[-1,:]
        self.q0 = self.q[-1, :]
        
        self.xdes = np.array(xdes)
        
        
        self.xv = x[-1, :]
        self.tv = t[-1]
        self.e = -2
        self.ev = []
        self.event = 'v'
        self.xx = np.array([x[-1, :]])        
        self.xx, self.q0 = self.trans()
        self.xvlo = np.array([self.xv])
        self.x[-1,:] = self.xx
        self.q[-1, :] = self.q0
        
        self.dt = dt
        
        self.set_k = False

    def __call__(self):
        """
        executes hybrid system
        """
        self.t0 = self.t[-1]
        self.x0 = self.x[-1,:]
        self.q0 = self.q[-1, :]
        
        self.xx = integrate.odeint(self.dyn, self.x0, np.array([0,self.dt]))                                   
        
        self.e = -1
        self.ev = list(self.evts())      
        self.ev = np.array([self.ev[i](self.t0 + self.dt, \
        self.xx[-1,:]) for i in range(len(self.ev))])
        
        if (self.ev>0).any():
            self.e = list(self.ev>0).index(True)
            self.trefined , self.xxrefined = self.refine(self.evts()[self.e])
            self.xx = np.array([self.xxrefined])
            self.xx, self.q0 = self.trans()
            self.WhichEvt()
        
        if self.e!=0: self.event = ''

        self.x = np.append(self.x, [self.xx[-1,:]], axis=0)
        self.q = np.append(self.q, [self.q0], axis=0)
        if self.e >= 0:
            self.t = np.append(self.t, [self.trefined], axis=0)
        else:
            self.t = np.append(self.t, [self.t0 + self.dt], axis=0)
        if self.event=='v':
            self.xvlo = np.append(self.xvlo, [self.xv], axis=0)

        return None    
    
    def refine( self, evtFun, tol=1e-12, *args ):
        """Find the time at which the trajectory is a root of the
           function evtFun.
           
           The original code is part of the package 'integro', which is adopted
           here for notation consistency.
        
           evtFun is a function taking (t,y,*args) and returning a real number.
           refine() uses trajectory patch to find the "time"
           at which evtFun(t,y(t),*args)==0.
           
           This code requires that the signs of evtFun in the beginning and
           end of the patch are opposite. It finds the root by bisection 
           -- a binary search.
           
           returns t,y --the end time and state; t is correct to within +/- tol
        """
        t0,t1 = self.t0, self.t0 + self.dt
        for k in xrange(4):
            y = self.x0
            f0 = evtFun(t0,y, *args)
            y = integrate.odeint(self.dyn, self.x0, \
            np.array([0, np.abs(t1 - t0)]))[-1, :]
            f1 = evtFun(t1,y, *args)
            if f0*f1 <= 0:
                break
            t0,t1 = t0-(t1-t0)/2,t1+(t1-t0)/2
            print "WARNING: function did not change sign -- extrapolating order ",k
#        print 't0,f0,t1,f1', t0,f0,t1,f1
        if f1<0:
            t0,f0,t1,f1 = t1,f1,t0,f0
        ypre = self.x0
        while abs(t1-t0)>tol:
            t = (t1+t0)/2.
            y = integrate.odeint(self.dyn, ypre, \
            np.array([0, np.abs(t1 - t0)/2]))[-1, :]
            f = evtFun(t,y, *args )
            if f==0:
                break
            if f>0:
                t1,f1 = t,f
            else:
                t0,f0 = t,f
                ypre = y
        return (t1+t0)/2, y  
    
    def WhichEvt(self):   
            if self.q0[0]==0:
                self.xv = self.xx[-1,:]
                self.tv = self.trefined
                self.event = 'v'
            elif self.q0[0]==1:
                self.xt = self.xx[-1,:]
                self.tt = self.trefined
                self.event = 't'
            elif self.q0[0]==2:
                self.xb = self.xx[-1,:]
                self.tb = self.trefined
                self.event = 'b'
            elif self.q0[0]==3:
                self.xl = self.xx[-1,:]
                self.tl = self.trefined
                self.event = 'l'
            return None
    
    def dyn(self, x, t):
        """
        .dyn  evaluates system dynamics
        """
        q,m,eta0,k,g,beta,foot1,foot2, b = self.q0
       
        # single-stance
        if q == 0 or q == 3:        
            q1, q2, dq1, dq2 = x
            if q == 0: c1x = foot1
            if q == 3: c1x = foot2
            c1y = 0
            sqrL1 = (-c1x + q1)**2 + (-c1y + q2)**2
            L1 = (sqrL1)**0.5
            dL1 = 1/2./L1*(2*dq1*(q1 - c1x) + 2*dq2*q2)
            dx = np.array([dq1, dq2, 1/m*(k*(eta0 - L1) - b*dL1)*(q1 - c1x)/L1,
            -g + 1/m*(k*(eta0 - L1) - b*dL1)*(q2 - c1y)/L1]) 

        # double-stance
        elif q == 1 or q == 2:        
            q1, q2, dq1, dq2 = x
            c1x = foot1; c1y = 0
            c2x = foot2; c2y = 0
            sqrL1 = (-c1x + q1)**2 + (-c1y + q2)**2
            L1 = (sqrL1)**0.5
            sqrL1 = (-c1x + q1)**2 + (-c1y + q2)**2
            sqrL2 = (-c2x + q1)**2 + (-c2y + q2)**2
            L2 = (sqrL2)**0.5
            dL1 = 1/2./L1*(2*dq1*(q1 - c1x) + 2*dq2*q2)
            dL2 = 1/2./L2*(2*dq1*(q1 - c2x) + 2*dq2*q2)
            ddq1 = 1/m*(k*(eta0 - L1) - b*dL1)*(q1 - c1x)/L1 +\
            1/m*(k*(eta0 - L2) - b*dL2)*(q1 - c2x)/L2 
            ddq2 = -g + 1/m*(k*(eta0 - L1) - b*dL1)*(q2 - c1y)/L1 + \
            1/m*(k*(eta0 - L2) - b*dL2)*(q2 -c2y)/L2
            dx = np.array([dq1, dq2, ddq1, ddq2])

        return dx

    

    def trans(self):
        """
        .trans  transition between discrete modes
        """
        q,m,eta0,k,g,beta,foot1,foot2, b = self.q0.copy()
        x = self.xx.copy()
        e = self.e
        
        # DS compression
        if q == 1:
            if any([e==i for i in range(1, len(self.ev))]): q = -1
            #bottom event
            elif e == 0:
                if self.set_k:
                    xb, yb, xdb, ydb = x
                    xv, yv, xdv, ydv = self.xdes
                    L1b = np.sqrt((xb - foot1)**2 + (yb)**2)
                    L2b = np.sqrt((xb - foot2)**2 + (yb)**2)
                    k = -m*(2.0*g*yb - 2.0*g*yv + xdb**2 + ydb**2 \
                    - xdv**2 - ydv**2)/ ((eta0 - L1b)**2 + \
                    (eta0 - L2b)**2 - (eta0 - yv)**2)
                    
                q = 2
                
        # DS decompression
        elif q == 2:
            if any([e==i for i in range(1, len(self.ev))]): q = -1
            elif e == 0: q = 3 #(liftoff event)

        # SS after vlo
        elif q == 0:
            if any([e==i for i in range(1, len(self.ev))]): q = -1
            elif e == 0: foot2 = x[-1, 0] + eta0*np.cos(beta);q = 1 #(Touchdown)
            elif e == -2 and x[-1, 1]<eta0*np.sin(beta):  q = -1
           
        # SS before vlo 
        elif q == 3:
            if any([e==i for i in range(1, len(self.ev))]):q = -1
            elif e == 0: #(VLO)
                if x[-1, 1]<eta0*np.sin(beta):  q = -1
                else:auxfoot = foot1; foot1 = foot2; foot2 = auxfoot; q = 0
        
        x = np.array(x)
        qq = np.array([q,m,eta0,k,g,beta,foot1, foot2, b])

        return x, qq


    def evts(self):
        """
        .evts  returns event functions for given hybrid domain
        """
        m,eta0,k,g,beta,foot1,foot2, b = self.q0[1:]

        # Bottom
        if self.q0[0] == 1:
            vb = (foot1 + foot2)/2.
            return [lambda t,x : ((x[0] - vb)*x[2] + x[1]*x[3])/ 
            np.sqrt((x[0]-vb)**2 + x[1]**2),
                    lambda t,x : -x[1],
                    lambda t,x : -x[2],
                    lambda t,x: np.sqrt((x[0]-foot1)**2 + x[1]**2) - eta0,
                    lambda t,x: np.sqrt((x[0]-foot2)**2 + x[1]**2) - eta0,
                    lambda t,x: np.sqrt((x[0]-foot1)**2 + x[1]**2) +\
                    np.sqrt((x[0]-foot2)**2 + x[1]**2)- 2*eta0,
                    lambda t,x: x[0] - foot2]
                    
        # Liftoff
        elif self.q0[0] == 2:
            return [lambda t,x : -(k*(eta0 - 
            np.sqrt((x[0]-foot1)**2 + x[1]**2)) - 
            b*(1/np.sqrt((x[0]-foot1)**2 + x[1]**2)*
            ((x[0]-foot1)*x[2] + x[1]*x[3]))),
                    lambda t,x : -x[1],
                    lambda t,x : -x[2],
                    lambda t,x: np.sqrt((x[0]-foot2)**2 + x[1]**2) - eta0,
                    lambda t,x: np.sqrt((x[0]-foot1)**2 + x[1]**2) +\
                    np.sqrt((x[0]-foot2)**2 + x[1]**2)- 2*eta0,
                    lambda t,x: x[0] - foot2]

        # Touchdown
        elif self.q0[0] == 0:
            # hip height
            h = eta0 * np.sin(beta)

            return [lambda t,x : -(x[1] - h),
                    lambda t,x : -x[1],
                    lambda t,x : -x[2],
                    lambda t,x: np.sqrt((x[0]-foot1)**2 + x[1]**2) - eta0]
        
        # VLO           
        elif self.q0[0] == 3:

            return [lambda t,x : x[0] - foot2,
                    lambda t,x : -x[1],
                    lambda t,x : -x[2],
                    lambda t,x: np.sqrt((x[0]-foot2)**2 + x[1]**2) - eta0]


    def get_phase(self):
        """
        Parameter
        ---------
        not needed.
        
        Return
        ------
        phase : string
                subphase ('av', 'sc', 'sd', 'bv')
        """
        if self.q0[0]==0:
            ph = 'av'
        elif self.q0[0]==1 :
            ph = 'sc'
        elif self.q0[0]==2:
            ph = 'sd'
        elif self.q0[0]==3:
            ph = 'bv'
        return ph
