# -*- coding: utf-8 -*-
"""
A Python Function computing an approximate mapping from any initial condition to
a particular event (e.g., touchdown, liftoff) at the end of certain mode of 
the Bipedal SLIP model.

The derivation of this map has been presented in 
"Approximate analytical solutions to the double-stance dynamics of 
the lossy Spring-Loaded Inverted Pendulum, 
by M. Shahbazi, U. Saranli, R. Babuska, G. Lopes" (under review).

Run "test.py" to see sample results.

Copyright (c) 2015 by Mohammad Shahbazi. 
doi:10.4121/uuid:fd95d86d-fa01-4faa-b6e5-eeb92345be8a
http://doi.org/10.4121/uuid:fd95d86d-fa01-4faa-b6e5-eeb92345be8a

This code can be freely used for any ends. 
If use of this code in part or in whole results in publication, proper citation must be
included in that publication. This code comes with no guarantees or support.

Mohammad Shahbazi, August 11, 2015, Delft, the Netherlands.
"""

def DBSLIP_Approx(x, q):

    from numpy import sin, cos, exp
    import numpy as np
#    from scipy.optimize import root, brute
    
    EnergyCorrect = True

    
    def dth_(t):
        if q in [1, 2] and xi!=0:
            return z0/r_(t)**2*\
            exp(-(z1*t + z2*exp(-xi*w0h*t)*cos(wd*t + phi - phi2)))

        else:
            return 3*w - 2*w*F/(L0*w0h**2) - 2*w*M*exp(-xi*w0h*t)*\
            cos(wd*t+phi)/L0
            
    
    def th_(t):
        if q in [1, 2] and xi!=0:
            return -z7/z5*exp(-(z5*t + z6)) + z10*exp(-(z5 + w0h*xi)*t)*\
            cos(wd*t + phi + phi3) + z7/z5*exp(-z6) + z11
            
        elif q in [1, 2] and xi==0:
            return z7*t + z8*sin(w0h*t + phi) + z9
        else:
            X = 3*w - 2*w*F/(L0*w0h**2)
            Y = 2*w*M/(L0*w0h)
            return th0 + X*t + Y*(exp(-xi*w0h*t)*cos(wd*t+phi-phi2)-\
            cos(phi-phi2))
            
            
       
    def r_(t):
        return M*exp(-xi*w0h*t)*cos(wd*t + phi) + F/w0h**2

    def dr_(t):
        return -M*w0h*exp(-xi*w0h*t)*cos(wd*t + phi + phi2)
        
    def c2p(x, y, dx, dy, origin):
        r = np.sqrt((x - origin)**2 + y**2)
        th = np.arctan2(-(x - origin), y)
        dth = - 1./r*(dy*sin(th) + dx*cos(th))
        dr = -dx*sin(th) + dy*cos(th)
        return r, th, dr, dth
        
    def p2c(r, th, dr, dth, origin):
        x = -r*sin(th) + origin
        y = r*cos(th)
        dx = -dr*sin(th) - r*dth*cos(th)
        dy = dr*cos(th) - r*dth*sin(th)
        return x, y, dx, dy
        
    def Params():
        w = p/mh/L0**2
        w0h = (k0/mh + 3*w**2)**0.5    
        F = -g0 + L0*k0/mh + 4.*L0*w**2
        xi = b/(2*mh*w0h)
        wd = w0h*np.sqrt(1 - xi**2)
        A = r0 - F/w0h**2
        B = (dr0 + xi*w0h*A)/wd               
        M = np.sqrt(A**2 + B**2)
        phi = np.arctan2(-B, A)
        phi2 = np.arctan2(-np.sqrt(1 - xi**2), xi)
        z1 = bt*(3*L0 - 2*F/w0h**2)/(mh*L0**3)
        z2 = -2*M*bt/(L0**3*mh*w0h)
        z0 = w*L0**2*exp(z2*cos(phi - phi2))
        z3 = -(wd*sin(phi - phi2) + w0h*xi*cos(phi - phi2))
        z4 = cos(phi - phi2)
        z5 = z1 + z2*z3 
        z6 = z2*z4
        z7 = z0*(3*L0 - 2*F/w0h**2)/L0**3
        z8 = -2*z0*M/(L0**3*w0h)
        z9 = th0 - z8*sin(phi)
        phi3 = np.arctan2(wd, (z5 + xi*w0h))
        z10 = z8*w0h*exp(-z6)/np.sqrt(z5**2 + 2*z5*w0h*xi + w0h**2)
        z11 = th0 - z10*cos(phi + phi3)
        z12 = -z10/np.sqrt(z5**2 + 2*z5*w0h*xi + w0h**2)
        z13 = M**2*wd**2*b/(4*xi*w0h)
        z14 = M**2*xi*w0h*b/2
        z15 = M**2*wd*b/4
        z16 = z13 + z14*cos(phi + phi2)**2 - z15*sin(2*phi + 2*phi2)
        phi4 = np.arctan2(wd, 2*z5 + xi*w0h)        
        z17 = 4*M*z0**2*bt*exp(-2*z6)/L0**5/\
        np.sqrt(wd**2 + (2*z5 + xi*w0h)**2)
        z18 = z0**2*(3*L0 - 4*F/w0h**2)*bt*np.exp(-2*z6)/2/L0**5/z5        
        return w, w0h, F, xi, wd, M, z0, z1, z2, z3,\
        z4, z5, z6, z7, z8, z9, z10, z11, z12, z13, z14, z15, z16, z17, z18,\
        phi, phi2, phi3, phi4

    def Predict():
        if q==0:
            def TDmap(x, L0, alpha, r_, th_):
                return r_(x)*np.cos(th_(x)) - L0*np.sin(alpha)
#            TDic = np.pi/wd/2          
#            TDargs = (L0, alpha, r_, th_)
#            TDdata = brute(TDmap, np.array([slice(0, .50, .001)]), \
#            args=TDargs, finish=None)
##            tTD = np.asscalar(TDdata.x)
#            tTD = TDdata
            tTD = 0
            while TDmap(tTD, L0, alpha, r_, th_)>=0 and tTD<=.5:
                tTD += 0.0001
            phTD = th_(tTD)
            phdTD = dth_(tTD)
            rTD = r_(tTD)
            rdTD = dr_(tTD)
            return tTD, rTD, phTD, rdTD, phdTD
        elif q==3:
            def V2map(x, th_):
                return (th_(x))
#            V2ic = np.pi/wd/2          
#            V2args = (th_, wd)
#            V2data = brute(V2map, np.array([slice(0, .5, .001)]),\
#            args=V2args, finish=None)
#            tV2 = V2data
            tV2 = 0
            while V2map(tV2, th_)>=0 and tV2<=.5:
                tV2 += 0.0001
            phV2 = th_(tV2)
            phdV2 = dth_(tV2)
            rV2 = r_(tV2)
            rdV2 = dr_(tV2)

            return tV2, rV2, phV2, rdV2, phdV2
        elif q in [1, 2]:
            def LOmap(x,r_,th_,dr_,dth_,L0,k0,b,foot1,origin):
#                return r_(x)**2 + c*r_(x)*np.cos(th_(x))+c**2/4. - 1**2 
#                return  (r_(x)*np.cos(th_(x)) + origin)**2 + \
#                (r_(x)*np.sin(th_(x)))**2 - L0**2
                x1, y1, dx1, dy1 = p2c(r_(x), th_(x), dr_(x), \
                dth_(x), origin)
                l1 = np.sqrt((x1 - foot1)**2 + y1**2)
                dl1 = 1/l1*((x1 - foot1)*dx1 + y1*dy1)
                aa = L0 - l1
                if b==0: return aa                
                else: return k0*(L0 - l1) - b*dl1
                
#            LOic = tLO              
#            LOargs = (r_, th_, dr_, dth_, L0_original, k0_original,\
#            b_original, foot1, origin)
#            LOdata = brute(LOmap, np.array([slice(0, .5, .001)]), \
#            args=LOargs, finish=None)
#            tLO = LOdata
            
            tLO = 0
            while LOmap(tLO,r_,th_,dr_,dth_,L0_original,k0_original,\
            b_original,foot1,origin)>=0 and tLO<=.5:
                tLO += .0001
            phLO = th_(tLO)
            phdLO = dth_(tLO)
            rLO = r_(tLO)
            rdLO = dr_(tLO)
#            print LOmap(tLO, 1, r_, th_)
            return tLO, rLO, phLO, rdLO, phdLO
        
    
    def Gravity_correct():
        rbar = F/w0h**2 - M/(w0h*te)*(z4 - exp(-xi*w0h*te)*\
        cos(wd*te + phi - phi2))
        out = te/2.*mh*g0*rbar*(sin(th0) + sin(the))
        return out
    
    def Torsional_correct():
        if xi==0:
            out = -kt*(z7/2.*te**2 + z9*te - z8/w0h*(cos(w0h*te + phi) - cos(phi)))
        else:
            out = -kt*((z7/z5*exp(-z6) + z11)*te - z7/z5**2*exp(-z6)*\
            (1 - exp(-z5*te)) + z12*(exp(-(z5 + w0h*xi)*te)\
            *cos(wd*te + phi + 2*phi3) - cos(phi + 2*phi3)))
        return out
        
    def ATSLIP():
#        mu1 = (c/2.)**2
#        mu2 = (mh*g0/2/k0)**2
#        coef = [1, -2*L0, L0**2 - (mu1 + mu2), 2*mu1*L0, -mu1*L0**2]
#        lcondidate = np.roots(coef)
#        Leq = max(filter(lambda x: x<L0 and x>0, lcondidate))
#        yeq = np.sqrt(Leq**2 - (c/2.)**2)
#        kr = mh*g0/(np.sqrt(L0**2 - c**2/4.) - yeq)
#        L0r = mh*g0/kr + yeq    
#        kt = 4*c**2*k0*L0*L0r**2/(c**2 + 4*L0r**2)**(3/2.)
#        br = (2*b*L0r**2)/(L0r**2 + c**2/4.)
#        bt = (b*c**2/2.*L0r**2)/(L0r**2 + c**2/4.)
#        th_rest = np.pi/2.
        
        L0r = np.sqrt(L0**2 - c**2/4.)
        kr = 2*k0 - c**2/2./L0**2*k0
        kt = k0/2/L0**2*c**2*L0r**2
        br = 2*b/L0**2*(L0**2 - c**2./4)
        bt = .5*b/L0**2*c**2*(L0**2 - c**2./4)
        th_rest = 0
        return L0r, kr, br, th_rest, kt, bt
        
            
    def Ebr():
        return exp(-2*xi*w0h*te)*(-z13 - z14*cos(wd*te + phi + phi2)**2\
        + z15*sin(2*wd*te + 2*phi + 2*phi2)) + z16
        
    def Ebt():
        return z17*(cos(wd*te + phi + phi4)*exp(-(2*z5 + xi*w0h)*te) - \
        cos(phi + phi4)) + z18*(exp(-2*z5*te) - 1)

        
    def CorrectEnergy():
        if xi != 0:
            Eb = Ebr()
            if q in [1, 2]: Eb += Ebt()
        else: 
            Eb = 0
        
        if q in [1, 2]:
            x0, y0, dx0, dy0 = p2c(r0, th0, dr0, dth0, origin)
            L1 = np.sqrt((x0 - foot1)**2 + y0**2)
            L2 = np.sqrt((x0 - foot2)**2 + y0**2)
            E0 = 1/2.*mh*(dx0**2 + dy0**2) + mh*g0*y0 +\
            1/2.*k0_original*(L0_original - L1)**2 + \
            1/2.*k0_original*(L0_original - L2)**2
            
        else: E0 = 1/2.*mh*(dr0**2 + (r0*dth0)**2) + mh*g0*r0*np.cos(th0)\
        + 1/2.*k0*(L0 - r0)**2
        if q in [1, 2]: 
            xe, ye, dxe, dye = p2c(re, the, dre, dthe, origin)
            L1 = np.sqrt((xe - foot1)**2 + ye**2)
            L2 = np.sqrt((xe - foot2)**2 + ye**2)
            Eehat = mh*g0*ye +\
            1/2.*k0_original*(L0_original - L1)**2 + \
            1/2.*k0_original*(L0_original - L2)**2
        else: Eehat = mh*g0*re*np.cos(the) + 1/2.*k0*(L0 - re)**2
       
        ve = np.sqrt((E0 - Eehat - Eb)*2/mh)
        dthehat = np.sign(dthe)*np.sqrt(np.abs(ve**2 - dre**2))/re        
        return dthehat
        
    #######################################################
    #######################################################        
    x0, y0, dx0, dy0 = np.array(x).copy()
    q,mh,L0,k0,g0,alpha,foot1,foot2, b = np.array(q).copy()
    
    if q == 0: origin=foot1; th_rest, kt, bt = 0, 0, 0
    elif q == 3: origin=foot2; th_rest, kt, bt = 0, 0, 0
    elif q in [1, 2]:
        L0_original = L0; k0_original = k0; b_original = b
        c = np.abs(foot1 - foot2)
        origin=(foot1 + foot2)/2.
        L0, k0, b, th_rest, kt, bt = ATSLIP()        
    
    r0, th0, dr0, dth0 = c2p(x0, y0, dx0, dy0, origin)

        
    p = dth0*mh*r0**2
    
    
    #  Frist Try:
    w, w0h, F, xi, wd, M, z0, z1, z2, z3,\
    z4, z5, z6, z7, z8, z9, z10, z11, z12, z13, z14, z15, z16, z17, z18,\
    phi, phi2, phi3, phi4 = Params()
    
    te, re, the, dre, dthe = Predict()
    
    p += Gravity_correct()
    
    if q in [1, 2]: p += Torsional_correct()
    
    # Second Try:    
    w, w0h, F, xi, wd, M, z0, z1, z2, z3,\
    z4, z5, z6, z7, z8, z9, z10, z11, z12, z13, z14, z15, z16, z17, z18,\
    phi, phi2, phi3, phi4 = Params()   
        
    te, re, the, dre, dthe = Predict()
    
#    print 'dthe before', dthe    
    if EnergyCorrect: dthe = CorrectEnergy()
#    print 'dthe after', dthe
    

    time = np.linspace(0, te, 200)
    r = r_(time)
#    Lmin = min(r)
    
    xe, ye, dxe, dye = p2c(re, the, dre, dthe, origin)
    
    x, y, dx, dy = p2c(r_(time), th_(time), dr_(time), dth_(time), origin)
    x[-1], y[-1], dx[-1], dy[-1] = xe, ye, dxe, dye
    
    if q in [1, 2]: return time, x, y, dx, dy
    
    else: return time, x, y, dx, dy
